@@ -15,7 +15,6 @@
 #include "soup-connection.h"
 #include "soup-server-message-private.h"
 #include "soup-message-headers-private.h"
-#include "soup-socket.h"
 #include "soup-uri-utils-private.h"
 
 /**
@@ -38,14 +37,11 @@
 struct _SoupServerMessage {
         GObject             parent;
 
-        SoupSocket         *sock;
-        GSocket            *gsock;
+        SoupServerConnection *conn;
         SoupAuthDomain     *auth_domain;
         char               *auth_user;
 
-        GSocketAddress     *remote_addr;
         char               *remote_ip;
-        GSocketAddress     *local_addr;
 
         const char         *method;
         SoupHTTPVersion     http_version;
@@ -62,7 +58,7 @@ struct _SoupServerMessage {
         SoupMessageBody    *response_body;
         SoupMessageHeaders *response_headers;
 
-        SoupServerMessageIOData *io_data;
+        SoupServerMessageIO *io_data;
 
         gboolean                 options_ping;
 
@@ -87,6 +83,7 @@ enum {
         GOT_CHUNK,
         GOT_BODY,
 
+        CONNECTED,
         DISCONNECTED,
         FINISHED,
 
@@ -123,18 +120,13 @@ soup_server_message_finalize (GObject *object)
 {
         SoupServerMessage *msg = SOUP_SERVER_MESSAGE (object);
 
-        soup_server_message_io_data_free (msg->io_data);
-
         g_clear_object (&msg->auth_domain);
         g_clear_pointer (&msg->auth_user, g_free);
-        g_clear_object (&msg->remote_addr);
-        g_clear_object (&msg->local_addr);
 
-        if (msg->sock) {
-                g_signal_handlers_disconnect_by_data (msg->sock, msg);
-                g_object_unref (msg->sock);
+        if (msg->conn) {
+                g_signal_handlers_disconnect_by_data (msg->conn, msg);
+                g_object_unref (msg->conn);
         }
-        g_clear_object (&msg->gsock);
         g_clear_pointer (&msg->remote_ip, g_free);
 
         g_clear_pointer (&msg->uri, g_uri_unref);
@@ -327,6 +319,20 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
                               NULL, NULL,
                               NULL,
                               G_TYPE_NONE, 0);
+        /**
+         * SoupServerMessage::connected:
+         * @msg: the message
+         *
+         * Emitted when the @msg's socket is connected and the TLS handshake completed.
+         */
+        signals[CONNECTED] =
+                g_signal_new ("connected",
+                              G_OBJECT_CLASS_TYPE (object_class),
+                              G_SIGNAL_RUN_LAST,
+                              0,
+                              NULL, NULL,
+                              NULL,
+                              G_TYPE_NONE, 0);
 
         /**
          * SoupServerMessage::disconnected:
@@ -403,15 +409,24 @@ soup_server_message_class_init (SoupServerMessageClass *klass)
 }
 
 static void
-socket_disconnected (SoupServerMessage *msg)
+connection_connected (SoupServerMessage *msg)
 {
+        g_assert (!msg->io_data);
+        msg->io_data = soup_server_connection_get_io_data (msg->conn);
+        g_signal_emit (msg, signals[CONNECTED], 0);
+}
+
+static void
+connection_disconnected (SoupServerMessage *msg)
+{
+        msg->io_data = NULL;
         g_signal_emit (msg, signals[DISCONNECTED], 0);
 }
 
 static gboolean
-socket_accept_certificate (SoupServerMessage    *msg,
-                           GTlsCertificate      *tls_certificate,
-                           GTlsCertificateFlags *tls_errors)
+connection_accept_certificate (SoupServerMessage    *msg,
+                               GTlsCertificate      *tls_certificate,
+                               GTlsCertificateFlags *tls_errors)
 {
 	gboolean accept = FALSE;
 
@@ -436,33 +451,34 @@ soup_server_message_set_tls_peer_certificate (SoupServerMessage   *msg,
 }
 
 static void
-re_emit_tls_certificate_changed (SoupServerMessage *msg,
-                                 GParamSpec        *pspec,
-                                 SoupSocket        *sock)
+re_emit_tls_certificate_changed (SoupServerMessage    *msg,
+                                 GParamSpec           *pspec,
+                                 SoupServerConnection *conn)
 {
         soup_server_message_set_tls_peer_certificate (msg,
-                                                      soup_socket_get_tls_certificate (sock),
-                                                      soup_socket_get_tls_certificate_errors (sock));
+                                                      soup_server_connection_get_tls_peer_certificate (conn),
+                                                      soup_server_connection_get_tls_peer_certificate_errors (conn));
 }
 
 SoupServerMessage *
-soup_server_message_new (SoupSocket *sock)
+soup_server_message_new (SoupServerConnection *conn)
 {
         SoupServerMessage *msg;
 
         msg = g_object_new (SOUP_TYPE_SERVER_MESSAGE, NULL);
-        msg->sock = g_object_ref (sock);
-        msg->gsock = soup_socket_get_gsocket (sock);
-        if (msg->gsock)
-                g_object_ref (msg->gsock);
+        msg->conn = g_object_ref (conn);
+        msg->io_data = soup_server_connection_get_io_data (msg->conn);
 
-        g_signal_connect_object (sock, "disconnected",
-                                 G_CALLBACK (socket_disconnected),
+        g_signal_connect_object (conn, "connected",
+                                 G_CALLBACK (connection_connected),
                                  msg, G_CONNECT_SWAPPED);
-        g_signal_connect_object (sock, "accept-certificate",
-                                 G_CALLBACK (socket_accept_certificate),
+        g_signal_connect_object (conn, "disconnected",
+                                 G_CALLBACK (connection_disconnected),
                                  msg, G_CONNECT_SWAPPED);
-        g_signal_connect_object (sock, "notify::tls-certificate",
+        g_signal_connect_object (conn, "accept-certificate",
+                                 G_CALLBACK (connection_accept_certificate),
+                                 msg, G_CONNECT_SWAPPED);
+        g_signal_connect_object (conn, "notify::tls-certificate",
                                  G_CALLBACK (re_emit_tls_certificate_changed),
                                  msg, G_CONNECT_SWAPPED);
 
@@ -478,10 +494,10 @@ soup_server_message_set_uri (SoupServerMessage *msg,
         msg->uri = soup_uri_copy_with_normalized_flags (uri);
 }
 
-SoupSocket *
-soup_server_message_get_soup_socket (SoupServerMessage *msg)
+SoupServerConnection *
+soup_server_message_get_connection (SoupServerMessage *msg)
 {
-        return msg->sock;
+        return msg->conn;
 }
 
 void
@@ -501,6 +517,9 @@ soup_server_message_set_auth (SoupServerMessage *msg,
 gboolean
 soup_server_message_is_keepalive (SoupServerMessage *msg)
 {
+        if (msg->http_version == SOUP_HTTP_2_0)
+                return TRUE;
+
         if (msg->status_code == SOUP_STATUS_OK && msg->method == SOUP_METHOD_CONNECT)
                 return TRUE;
 
@@ -536,19 +555,76 @@ soup_server_message_is_keepalive (SoupServerMessage *msg)
 }
 
 void
-soup_server_message_set_io_data (SoupServerMessage       *msg,
-                                 SoupServerMessageIOData *io)
+soup_server_message_read_request (SoupServerMessage        *msg,
+                                  SoupMessageIOCompletionFn completion_cb,
+                                  gpointer                  user_data)
 {
-        soup_server_message_io_data_free (msg->io_data);
-        msg->io_data = io;
+        soup_server_message_io_read_request (msg->io_data, msg, completion_cb, user_data);
 }
 
-SoupServerMessageIOData *
+SoupServerMessageIO *
 soup_server_message_get_io_data (SoupServerMessage *msg)
 {
         return msg->io_data;
 }
 
+/**
+ * soup_server_message_pause:
+ * @msg: a SoupServerMessage
+ *
+ * Pauses I/O on @msg.
+ *
+ * This can be used when you need to return from the server handler without
+ * having the full response ready yet. Use [method@ServerMessage.unpause] to
+ * resume I/O.
+ *
+ * Since: 3.2
+ */
+void
+soup_server_message_pause (SoupServerMessage *msg)
+{
+        g_return_if_fail (SOUP_IS_SERVER_MESSAGE (msg));
+        g_return_if_fail (msg->io_data != NULL);
+
+        soup_server_message_io_pause (msg->io_data, msg);
+}
+
+/**
+ * soup_server_message_unpause:
+ * @msg: a SoupServerMessage
+ *
+ * Resumes I/O on @msg.
+ *
+ * Use this to resume after calling [method@ServerMessage.pause], or after
+ * adding a new chunk to a chunked response. I/O won't actually resume until you
+ * return to the main loop.
+ *
+ * Since: 3.2
+ */
+void
+soup_server_message_unpause (SoupServerMessage *msg)
+{
+        g_return_if_fail (SOUP_IS_SERVER_MESSAGE (msg));
+        g_return_if_fail (msg->io_data != NULL);
+
+        soup_server_message_io_unpause (msg->io_data, msg);
+}
+
+gboolean
+soup_server_message_is_io_paused (SoupServerMessage *msg)
+{
+        return msg->io_data && soup_server_message_io_is_paused (msg->io_data, msg);
+}
+
+void
+soup_server_message_finish (SoupServerMessage *msg)
+{
+        if (!msg->io_data)
+                return;
+
+        soup_server_message_io_finished (g_steal_pointer (&msg->io_data), msg);
+}
+
 void
 soup_server_message_cleanup_response (SoupServerMessage *msg)
 {
@@ -930,7 +1006,7 @@ soup_server_message_get_socket (SoupServerMessage *msg)
 {
         g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), NULL);
 
-        return msg->gsock;
+        return soup_server_connection_get_socket (msg->conn);
 }
 
 /**
@@ -949,14 +1025,7 @@ soup_server_message_get_remote_address (SoupServerMessage *msg)
 {
         g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), NULL);
 
-        if (msg->remote_addr)
-                return msg->remote_addr;
-
-        msg->remote_addr = msg->gsock ?
-                g_socket_get_remote_address (msg->gsock, NULL) :
-                G_SOCKET_ADDRESS (g_object_ref (soup_socket_get_remote_address (msg->sock)));
-
-        return msg->remote_addr;
+        return soup_server_connection_get_remote_address (msg->conn);
 }
 
 /**
@@ -975,14 +1044,7 @@ soup_server_message_get_local_address (SoupServerMessage *msg)
 {
         g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), NULL);
 
-        if (msg->local_addr)
-                return msg->local_addr;
-
-        msg->local_addr = msg->gsock ?
-                g_socket_get_local_address (msg->gsock, NULL) :
-                G_SOCKET_ADDRESS (g_object_ref (soup_socket_get_local_address (msg->sock)));
-
-        return msg->local_addr;
+        return soup_server_connection_get_local_address (msg->conn);
 }
 
 /**
@@ -1001,21 +1063,15 @@ soup_server_message_get_remote_host (SoupServerMessage *msg)
 {
         g_return_val_if_fail (SOUP_IS_SERVER_MESSAGE (msg), NULL);
 
-        if (msg->remote_ip)
-                return msg->remote_ip;
-
-        if (msg->gsock) {
-                GSocketAddress *addr = soup_server_message_get_remote_address (msg);
+        if (!msg->remote_ip) {
+                GSocketAddress *addr = soup_server_connection_get_remote_address (msg->conn);
                 GInetAddress *iaddr;
 
                 if (!addr || !G_IS_INET_SOCKET_ADDRESS (addr))
                         return NULL;
+
                 iaddr = g_inet_socket_address_get_address (G_INET_SOCKET_ADDRESS (addr));
                 msg->remote_ip = g_inet_address_to_string (iaddr);
-        } else {
-                GInetSocketAddress *addr = G_INET_SOCKET_ADDRESS (soup_socket_get_remote_address (msg->sock));
-                GInetAddress *inet_addr = g_inet_socket_address_get_address (addr);
-                msg->remote_ip = g_inet_address_to_string (inet_addr);
         }
 
         return msg->remote_ip;
@@ -1047,16 +1103,9 @@ soup_server_message_steal_connection (SoupServerMessage *msg)
         GIOStream *stream;
 
         g_object_ref (msg);
-        stream = soup_server_message_io_steal (msg);
-        if (stream) {
-                g_object_set_data_full (G_OBJECT (stream), "GSocket",
-                                        soup_socket_steal_gsocket (msg->sock),
-                                        g_object_unref);
-        }
-
-        g_signal_handlers_disconnect_by_data (msg, msg->sock);
-
-        socket_disconnected (msg);
+        stream = soup_server_connection_steal (msg->conn);
+        g_signal_handlers_disconnect_by_data (msg, msg->conn);
+        connection_disconnected (msg);
         g_object_unref (msg);
 
         return stream;
